use std::alloc::{alloc, dealloc, Layout};

extern crate wee_alloc;

#[global_allocator]
static ALLOC: wee_alloc::WeeAlloc = wee_alloc::WeeAlloc::INIT;

// webassembly version of all the wordle logic
// because the javascript version is slow as hell on firefox
use std::os::raw::c_char;

static mut LOOKUP: Option<&'static mut [u8]> = None;
static mut STRINGS: Option<&'static mut [u8]> = None;
static mut STR_IDXS: Option<&'static mut [*const c_char]> = None;

fn unwrap_or_abort<T>(v: Option<T>) -> T {
    match v {
        Some(v) => v,
        None => std::process::abort(),
    }
}

unsafe fn alloc_ary<T>(sz: usize) -> &'static mut [T] {
    let layout = unwrap_or_abort(Layout::from_size_align(
            sz*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    let ptr = alloc(layout) as *mut T;
    std::slice::from_raw_parts_mut(ptr, sz)
}

unsafe fn dealloc_ary<T>(v: &'static mut [T]) {
    let layout = unwrap_or_abort(Layout::from_size_align(
            v.len()*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    dealloc(v.as_ptr() as *mut u8, layout);
}


#[no_mangle]
pub extern fn init_strings(sz: usize) -> *mut c_char {
    unsafe {
        if let Some(v) = &mut STRINGS {
            dealloc_ary(v);
            STRINGS = None;
        }
        STRINGS = Some(alloc_ary(sz));
        unwrap_or_abort(STRINGS.as_deref_mut().map(|v| v.as_mut_ptr() as *mut c_char))
    }
}

#[no_mangle]
pub extern fn init_idx() -> usize {
    // TODO
    0
}

#[no_mangle]
pub extern fn init_lookup() -> *mut c_char {
    unsafe {
        let num_strings = STR_IDXS.as_ref().map(|idxs| idxs.len()).unwrap_or(0);
        if let Some(v) = &mut LOOKUP {
            dealloc_ary(v);
            STRINGS = None;
        }

        LOOKUP = Some(alloc_ary(num_strings*num_strings));
        unwrap_or_abort(
            LOOKUP.as_deref_mut().map(|v| v.as_mut_ptr() as *mut c_char))
    }
}

#[no_mangle]
pub extern fn calc_match(guess: *const c_char, reference: *const c_char) -> u8 {
    let mut unmatched = vec![0u8; 26];
    let mut ret = 0;
    let mut matched = 0;

    let mut mul = 1;
    for i in 0..5 {
        unsafe {
            if *(guess.offset(i)) == *(reference.offset(i)) {
                ret += 1*mul;
                matched |= 1 << i;
            } else {
                unmatched[*(reference.offset(i)) as usize - 'a' as usize] += 1;
            }
        }
        mul *= 3;
    }

    mul = 1;
    for i in 0..5 {
        if (matched & (1 << i)) == 0 {
            let idx = unsafe { *(guess.offset(i)) as usize - ('a' as usize) };
            if unmatched[idx] > 0 {
                ret += 2*mul;
                unmatched[idx] -= 1;
            }
        }
        mul *= 3;
    }

    ret
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() {
        let result = add(2, 2);
        assert_eq!(result, 4);
    }
}
